import functools
import sys
import weakref

if sys.version_info < (3, 9):
    memoize = functools.lru_cache(maxsize=None)
else:
    memoize = functools.cache

_memoize_weak_cache = weakref.WeakKeyDictionary()


def memoize_weak(fn):
    @functools.wraps(fn)
    def _wrapper(target, *args, **kwargs):
        # We need to avoid doing more than one read from each WeakKeyDict in
        # case our argument is not the same instance as the key in the cache.
        # Using this decorator for arguments or callables with value semantics
        # is a little pointless, but that's no excuse for blowing up.
        try:
            fn_cache = _memoize_weak_cache[fn]
        except KeyError:
            fn_cache = weakref.WeakKeyDictionary()
            _memoize_weak_cache[fn] = fn_cache

        try:
            closure = fn_cache[target]
        except KeyError:
            closure = memoize(
                lambda *args, **kwargs: fn(target, *args, **kwargs)
            )
            fn_cache[target] = closure

        return closure(*args, **kwargs)

    return _wrapper


def sort_key_from_iter(values):
    index = {statement: index for index, statement in enumerate(values)}
    key = lambda value: index[value]
    return key



import ast
import functools
import sys


@functools.singledispatch
def get_bindings(node):
    raise NotImplementedError(
        f"could not find bindings for unsupported node {node!r} "
        f"at line {node.lineno}, column: {node.col_offset}"
    )


def _get_bindings_from_arguments(args):
    for arg in args.posonlyargs:
        if arg.annotation is None:
            continue
        yield from get_bindings(arg.annotation)

    for arg in args.args:
        if arg.annotation is None:
            continue
        yield from get_bindings(arg.annotation)

    if args.vararg is not None and args.vararg.annotation is not None:
        yield from get_bindings(args.vararg.annotation)

    for arg in args.kwonlyargs:
        if arg.annotation is None:
            continue
        yield from get_bindings(arg.annotation)

    if args.kwarg is not None and args.kwarg.annotation is not None:
        yield from get_bindings(args.kwarg.annotation)

    for default in args.defaults:
        yield from get_bindings(default)

    for default in args.kw_defaults:
        if default is None:
            continue

        yield from get_bindings(default)


@get_bindings.register(ast.FunctionDef)
@get_bindings.register(ast.AsyncFunctionDef)
def _get_bindings_for_function_def(node):
    """
    ..code:: python

        FunctionDef(
            identifier name,
            arguments args,
            stmt* body,
            expr* decorator_list,
            expr? returns,
            string? type_comment,
        )

    ..code:: python

        AsyncFunctionDef(
            identifier name,
            arguments args,
            stmt* body,
            expr* decorator_list,
            expr? returns,
            string? type_comment,
        )

    """
    for decorator in node.decorator_list:
        yield from get_bindings(decorator)

    yield node.name

    yield from _get_bindings_from_arguments(node.args)

    if node.returns is not None:
        yield from get_bindings(node.returns)


@get_bindings.register(ast.ClassDef)
def _get_bindings_for_class_def(node):
    """
    ..code:: python

        ClassDef(
            identifier name,
            expr* bases,
            keyword* keywords,
            stmt* body,
            expr* decorator_list,
        )
    """
    for decorator in node.decorator_list:
        yield from get_bindings(decorator)

    for base in node.bases:
        yield from get_bindings(base)

    for keyword in node.keywords:
        yield from get_bindings(keyword.value)

    yield node.name


@get_bindings.register(ast.Return)
def _get_bindings_for_return(node):
    """
    ..code:: python

        Return(expr? value)

    """
    return
    yield


@get_bindings.register(ast.Delete)
def _get_bindings_for_delete(node):
    """
    ..code:: python

        Delete(expr* targets)
    """
    return
    yield


@functools.singledispatch
def _flatten_target(node):
    raise NotImplementedError(f"not implemented for {node!r}")


@_flatten_target.register(ast.Name)
def _flatten_target_name(node):
    assert isinstance(node.ctx, ast.Store)
    yield node.id


@_flatten_target.register(ast.Starred)
def _flatten_target_starred(node):
    assert isinstance(node.ctx, ast.Store)
    yield from _flatten_target(node.value)


@_flatten_target.register(ast.Tuple)
def _flatten_target_tuple(node):
    assert isinstance(node.ctx, ast.Store)
    for element in node.elts:
        yield from _flatten_target(element)


@_flatten_target.register(ast.Subscript)
def _flatten_target_subscript(node):
    assert isinstance(node.ctx, ast.Store)
    return
    yield


@_flatten_target.register(ast.Attribute)
def _flatten_target_attribute(node):
    """
    ..code:: python

        Attribute(expr value, identifier attr, expr_context ctx)
    """
    assert isinstance(node.ctx, ast.Store)
    return
    yield


@_flatten_target.register(ast.List)
def _flatten_target_list(node):
    assert isinstance(node.ctx, ast.Store)
    for element in node.elts:
        yield from _flatten_target(element)


@get_bindings.register(ast.Assign)
def _get_bindings_for_assign(node):
    """
    ..code:: python

        Assign(expr* targets, expr value, string? type_comment)
    """
    yield from get_bindings(node.value)

    for target in node.targets:
        yield from _flatten_target(target)


@get_bindings.register(ast.AugAssign)
def _get_bindings_for_aug_assign(node):
    """
    ..code:: python

        AugAssign(expr target, operator op, expr value)
    """
    yield from get_bindings(node.value)

    yield from _flatten_target(node.target)


@get_bindings.register(ast.AnnAssign)
def _get_bindings_for_ann_assign(node):
    """
    ..code:: python

        # 'simple' indicates that we annotate simple name without parens
        AnnAssign(expr target, expr annotation, expr? value, int simple)

    """
    yield from get_bindings(node.annotation)

    if node.value is not None:
        yield from get_bindings(node.value)

    yield from _flatten_target(node.target)


@get_bindings.register(ast.For)
@get_bindings.register(ast.AsyncFor)
def _get_bindings_for_for(node):
    """
    ..code:: python

        # use 'orelse' because else is a keyword in target languages
        For(
            expr target,
            expr iter,
            stmt* body,
            stmt* orelse,
            string? type_comment,
        )

    ..code:: python

        AsyncFor(
            expr target,
            expr iter,
            stmt* body,
            stmt* orelse,
            string? type_comment,
        )
    """
    yield from get_bindings(node.iter)

    yield from _flatten_target(node.target)

    for stmt in node.body:
        yield from get_bindings(stmt)

    for stmt in node.orelse:
        yield from get_bindings(stmt)


@get_bindings.register(ast.While)
def _get_bindings_for_while(node):
    """
    ..code:: python

        While(expr test, stmt* body, stmt* orelse)
    """
    yield from get_bindings(node.test)

    for stmt in node.body:
        yield from get_bindings(stmt)

    for stmt in node.orelse:
        yield from get_bindings(stmt)


@get_bindings.register(ast.If)
def _get_bindings_for_if(node):
    """
    ..code:: python

        If(expr test, stmt* body, stmt* orelse)
    """
    yield from get_bindings(node.test)

    for stmt in node.body:
        yield from get_bindings(stmt)

    for stmt in node.orelse:
        yield from get_bindings(stmt)


@get_bindings.register(ast.With)
@get_bindings.register(ast.AsyncWith)
def _get_bindings_for_with(node):
    """
    ..code:: python

        With(withitem* items, stmt* body, string? type_comment)

    ..code:: python

        AsyncWith(withitem* items, stmt* body, string? type_comment)
    """
    for item in node.items:
        yield from get_bindings(item.context_expr)

        if item.optional_vars:
            yield from _flatten_target(item.optional_vars)

    for stmt in node.body:
        yield from get_bindings(stmt)


@get_bindings.register(ast.Raise)
def _get_bindings_for_raise(node):
    """
    ..code:: python


        Raise(expr? exc, expr? cause)
    """
    if node.cause is not None:
        yield from get_bindings(node.cause)

    if node.exc is not None:
        yield from get_bindings(node.exc)


@get_bindings.register(ast.Try)
def _get_bindings_for_try(node):
    """
    ..code:: python

        Try(
            stmt* body,
            excepthandler* handlers,
            stmt* orelse,
            stmt* finalbody,
        )

        ExceptHandler(expr? type, identifier? name, stmt* body)
    """

    for stmt in node.body:
        yield from get_bindings(stmt)

    for handler in node.handlers:
        if handler.name:
            yield handler.name

        for stmt in handler.body:
            yield from get_bindings(stmt)

    for stmt in node.orelse:
        yield from get_bindings(stmt)

    for stmt in node.finalbody:
        yield from get_bindings(stmt)


@get_bindings.register(ast.Assert)
def _get_bindings_for_assert(node):
    """
    ..code:: python

        Assert(expr test, expr? msg)

    """
    yield from get_bindings(node.test)

    if node.msg is not None:
        yield from get_bindings(node.msg)


@get_bindings.register(ast.Import)
def _get_bindings_for_import(node):
    """
    ..code:: python

        Import(alias* names)
    """
    for name in node.names:
        if name.asname:
            yield name.asname
        else:
            root, *rest = name.name.split(".", 1)
            yield root


@get_bindings.register(ast.ImportFrom)
def _get_bindings_for_import_from(node):
    """
    ..code:: python

        ImportFrom(identifier? module, alias* names, int? level)

    """
    for name in node.names:
        yield name.asname if name.asname else name.name


@get_bindings.register(ast.Global)
def _get_bindings_for_global(node):
    """
    ..code:: python

        Global(identifier* names)
    """
    for name in node.names:
        yield name


@get_bindings.register(ast.Nonlocal)
def _get_bindings_for_non_local(node):
    """
    ..code:: python

        Nonlocal(identifier* names)
    """
    for name in node.names:
        yield name


@get_bindings.register(ast.Expr)
def _get_bindings_for_expr(node):
    """
    ..code:: python

        Expr(expr value)
    """
    yield from get_bindings(node.value)


@get_bindings.register(ast.Pass)
@get_bindings.register(ast.Break)
@get_bindings.register(ast.Continue)
def _get_bindings_for_control_flow(node):
    """
    ..code:: python

        Pass | Break | Continue

    """
    return
    yield


@get_bindings.register(ast.NamedExpr)
def _get_bindings_for_named_expr(node):
    """
    ..code:: python

        NamedExpr(expr target, expr value)
    """
    yield from get_bindings(node.value)
    yield from _flatten_target(node.target)


@get_bindings.register(ast.BoolOp)
def _get_bindings_for_bool_op(node):
    """
    ..code:: python

        # BoolOp() can use left & right?
        # expr
        BoolOp(boolop op, expr* values)
    """
    for expr in node.values:
        yield from get_bindings(expr)


@get_bindings.register(ast.BinOp)
def _get_bindings_for_bin_op(node):

    """
    ..code:: python

        BinOp(expr left, operator op, expr right)
    """
    yield from get_bindings(node.left)
    yield from get_bindings(node.right)


@get_bindings.register(ast.UnaryOp)
def _get_bindings_for_unary_op(node):
    """
    ..code:: python

        UnaryOp(unaryop op, expr operand)
    """
    yield from get_bindings(node.operand)


@get_bindings.register(ast.Lambda)
def _get_bindings_for_lambda(node):
    """
    ..code:: python

        Lambda(arguments args, expr body)
    """
    yield from _get_bindings_from_arguments(node.args)


@get_bindings.register(ast.IfExp)
def _get_bindings_for_if_exp(node):
    """
    ..code:: python

        IfExp(expr test, expr body, expr orelse)
    """
    yield from get_bindings(node.test)
    yield from get_bindings(node.body)
    yield from get_bindings(node.orelse)


@get_bindings.register(ast.Dict)
def _get_bindings_for_dict(node):
    """
    ..code:: python

        Dict(expr* keys, expr* values)
    """
    for key, value in zip(node.keys, node.values):
        if key is not None:
            yield from get_bindings(key)
        if value is not None:
            yield from get_bindings(value)


@get_bindings.register(ast.Set)
def _get_bindings_for_set(node):
    """
    ..code:: python

        Set(expr* elts)
    """
    for elt in node.elts:
        yield from get_bindings(elt)


@get_bindings.register(ast.ListComp)
def _get_bindings_for_list_comp(node):
    """
    ..code:: python

        comprehension = (expr target, expr iter, expr* ifs, int is_async)
        ListComp(expr elt, comprehension* generators)
    """
    for generator in node.generators:
        yield from get_bindings(generator.iter)
        yield from _flatten_target(generator.target)
        for condition in generator.ifs:
            yield from get_bindings(condition)

    yield from get_bindings(node.elt)


@get_bindings.register(ast.SetComp)
def _get_bindings_for_set_comp(node):
    """
    ..code:: python

        comprehension = (expr target, expr iter, expr* ifs, int is_async)
        SetComp(expr elt, comprehension* generators)
    """
    for generator in node.generators:
        yield from get_bindings(generator.iter)
        yield from _flatten_target(generator.target)
        for condition in generator.ifs:
            yield from get_bindings(condition)

    yield from get_bindings(node.elt)


@get_bindings.register(ast.DictComp)
def _get_bindings_for_dict_comp(node):
    """
    ..code:: python

        DictComp(expr key, expr value, comprehension* generators)
    """
    for generator in node.generators:
        yield from get_bindings(generator.iter)
        yield from _flatten_target(generator.target)
        for condition in generator.ifs:
            yield from get_bindings(condition)

    yield from get_bindings(node.key)
    yield from get_bindings(node.value)


@get_bindings.register(ast.GeneratorExp)
def _get_bindings_for_generator_exp(node):
    """
    ..code:: python

        GeneratorExp(expr elt, comprehension* generators)
    """
    for generator in node.generators:
        yield from get_bindings(generator.iter)
        yield from _flatten_target(generator.target)
        for condition in generator.ifs:
            yield from get_bindings(condition)

    yield from get_bindings(node.elt)


@get_bindings.register(ast.Await)
def _get_bindings_for_await(node):
    """
    ..code:: python

        # the grammar constrains where yield expressions can occur
        Await(expr value)
    """
    yield from get_bindings(node.value)


@get_bindings.register(ast.Yield)
def _get_bindings_for_yield(node):
    """
    ..code:: python

        Yield(expr? value)
    """
    if node.value is not None:
        yield from get_bindings(node.value)


@get_bindings.register(ast.YieldFrom)
def _get_bindings_for_yield_from(node):
    """
    ..code:: python

        YieldFrom(expr value)
    """
    yield from get_bindings(node.value)


@get_bindings.register(ast.Compare)
def _get_bindings_for_compare(node):
    """
    ..code:: python

        # need sequences for compare to distinguish between
        # x < 4 < 3 and (x < 4) < 3
        Compare(expr left, cmpop* ops, expr* comparators)
    """
    yield from get_bindings(node.left)
    for comparator in node.comparators:
        yield from get_bindings(comparator)


@get_bindings.register(ast.Call)
def _get_bindings_for_call(node):
    """
    ..code:: python

        keyword = (identifier? arg, expr value)
        Call(expr func, expr* args, keyword* keywords)
    """
    yield from get_bindings(node.func)

    for arg in node.args:
        yield from get_bindings(arg)

    for kwarg in node.keywords:
        yield from get_bindings(kwarg.value)


@get_bindings.register(ast.FormattedValue)
def _get_bindings_for_formatted_value(node):
    """
    ..code:: python

        FormattedValue(expr value, int? conversion, expr? format_spec)
    """
    yield from get_bindings(node.value)


@get_bindings.register(ast.JoinedStr)
def _get_bindings_for_joined_str(node):
    """
    ..code:: python

        JoinedStr(expr* values)
    """
    for value in node.values:
        yield from get_bindings(value)


@get_bindings.register(ast.Constant)
def _get_bindings_for_constant(node):
    """
    ..code:: python

        Constant(constant value, string? kind)
    """
    return
    yield


@get_bindings.register(ast.Attribute)
def _get_bindings_for_attribute(node):
    """
    ..code:: python

        Attribute(expr value, identifier attr, expr_context ctx)
    """
    yield from get_bindings(node.value)


@get_bindings.register(ast.Slice)
def _get_bindings_for_slice(node):
    """
    ..code:: python

        # can appear only in Subscript
        Slice(expr? lower, expr? upper, expr? step)

    """
    if node.lower is not None:
        yield from get_bindings(node.lower)

    if node.upper is not None:
        yield from get_bindings(node.upper)

    if node.step is not None:
        yield from get_bindings(node.step)


if sys.version_info < (3, 9):

    @get_bindings.register(ast.Index)
    def _get_bindings_for_index(node):
        """
        ..code:: python

            # can appear only in Subscript
            Index(expr value)
        """
        yield from get_bindings(node.value)


@get_bindings.register(ast.Subscript)
def _get_bindings_for_subscript(node):
    """
    ..code:: python

        Subscript(expr value, expr slice, expr_context ctx)
    """
    yield from get_bindings(node.value)
    yield from get_bindings(node.slice)


@get_bindings.register(ast.Starred)
def _get_bindings_for_starred(node):
    """
    ..code:: python

        Starred(expr value, expr_context ctx)
    """
    yield from get_bindings(node.value)


@get_bindings.register(ast.Name)
def _get_bindings_for_name(node):

    """
    ..code:: python

        Name(identifier id, expr_context ctx)
    """
    return
    yield


@get_bindings.register(ast.List)
def _get_bindings_for_list(node):

    """
    ..code:: python

        List(expr* elts, expr_context ctx)
    """
    for elt in node.elts:
        yield from get_bindings(elt)


@get_bindings.register(ast.Tuple)
def _get_bindings_for_tuple(node):
    """
    ..code:: python

        Tuple(expr* elts, expr_context ctx)

    """
    for elt in node.elts:
        yield from get_bindings(elt)



import sys

from ssort._builtins import MODULE_BUILTINS
from ssort._graphs import Graph
from ssort._statements import (
    statement_bindings,
    statement_method_requirements,
    statement_requirements,
)


class ResolutionError(Exception):
    def __init__(self, msg, *, unresolved):
        super().__init__(msg)
        self.unresolved = unresolved


def statements_graph(statements):
    """
    Returns a dictionary mapping from statements to lists of other statements.
    """
    # A dictionary mapping from names to the statements which bind them.
    scope = {}

    all_requirements = []
    resolved = {}

    for statement in statements:
        for requirement in statement_requirements(statement):
            all_requirements.append(requirement)

            # TODO error if requirement is not deferred.
            if requirement.name in scope:
                resolved[requirement] = scope[requirement.name]
                continue

            if requirement.name in MODULE_BUILTINS:
                resolved[requirement] = None
                continue

        for name in statement_bindings(statement):
            scope[name] = statement

    # Patch up dependencies that couldn't be resolved immediately.
    for requirement in all_requirements:
        if requirement in resolved:
            continue

        if requirement.name not in scope:
            continue

        resolved[requirement] = scope[requirement.name]

    if "*" in scope:
        sys.stderr.write("WARNING: can't determine dependencies on * import\n")

        for requirement in all_requirements:
            if requirement in resolved:
                continue

            resolved[requirement] = scope["*"]

    else:
        unresolved = [
            requirement
            for requirement in all_requirements
            if requirement not in resolved
        ]

        if unresolved:
            raise ResolutionError(
                "Could not resolve all requirements", unresolved=unresolved
            )

    graph = Graph()
    for statement in statements:
        graph.add_node(statement)

    for statement in statements:
        for requirement in statement_requirements(statement):
            if resolved[requirement] is not None:
                graph.add_dependency(statement, resolved[requirement])

    return graph


def class_statements_initialisation_graph(statements):
    scope = {}

    graph = Graph()

    for statement in statements:
        graph.add_node(statement)

        for requirement in statement_requirements(statement):
            if requirement.deferred:
                continue

            if requirement.name not in scope:
                continue

            graph.add_dependency(statement, scope[requirement.name])

        for name in statement_bindings(statement):
            scope[name] = statement

    return graph


def class_statements_runtime_graph(statements):
    scope = {}

    graph = Graph()

    for statement in statements:
        graph.add_node(statement)

        for name in statement_bindings(statement):
            scope[name] = statement

    for statement in statements:
        for name in statement_method_requirements(statement):
            if name in scope:
                graph.add_dependency(statement, scope[name])

    return graph



import ast
import dataclasses
import enum
import functools
import sys

from ssort._bindings import get_bindings
from ssort._builtins import CLASS_BUILTINS


class Scope(enum.Enum):
    LOCAL = "LOCAL"
    NONLOCAL = "NONLOCAL"
    GLOBAL = "GLOBAL"


@dataclasses.dataclass(frozen=True)
class Requirement:
    name: str
    lineno: int
    col_offset: int
    deferred: bool = False
    scope: Scope = Scope.LOCAL


@functools.singledispatch
def get_requirements(node):
    raise NotImplementedError(
        f"could not find requirements for unsupported node:  {node!r}"
        f"at line {node.lineno}, column: {node.col_offset}"
    )


def _get_scope_from_arguments(args):
    scope = set()
    scope.update(arg.arg for arg in args.posonlyargs)
    scope.update(arg.arg for arg in args.args)  # Arghhh.
    if args.vararg:
        scope.add(args.vararg.arg)
    scope.update(arg.arg for arg in args.kwonlyargs)
    if args.kwarg:
        scope.add(args.kwarg.arg)
    return scope


@get_requirements.register(ast.FunctionDef)
@get_requirements.register(ast.AsyncFunctionDef)
def _get_requirements_for_function_def(node):
    """
    ..code:: python

        FunctionDef(
            identifier name,
            arguments args,
            stmt* body,
            expr* decorator_list,
            expr? returns,
            string? type_comment,
        )

    ..code:: python

        AsyncFunctionDef(
            identifier name,
            arguments args,
            stmt* body,
            expr* decorator_list,
            expr? returns,
            string? type_comment,
        )

    """
    for decorator in node.decorator_list:
        yield from get_requirements(decorator)

    scope = _get_scope_from_arguments(node.args)

    requirements = []
    for statement in node.body:
        scope.update(get_bindings(statement))
        for requirement in get_requirements(statement):
            if not requirement.deferred:
                requirement = dataclasses.replace(requirement, deferred=True)
            requirements.append(requirement)

    for requirement in requirements:
        if requirement.scope == Scope.GLOBAL:
            yield requirement

        elif requirement.scope == Scope.NONLOCAL:
            yield dataclasses.replace(requirement, scope=Scope.LOCAL)

        else:
            if requirement.name not in scope:
                yield requirement


@get_requirements.register(ast.ClassDef)
def _get_requirements_for_class_def(node):
    """
    ..code:: python

        ClassDef(
            identifier name,
            expr* bases,
            keyword* keywords,
            stmt* body,
            expr* decorator_list,
        )
    """
    for decorator in node.decorator_list:
        yield from get_requirements(decorator)

    for base in node.bases:
        yield from get_requirements(base)

    scope = set(CLASS_BUILTINS)

    for statement in node.body:
        for stmt_dep in get_requirements(statement):
            if stmt_dep.deferred or stmt_dep.name not in scope:
                yield stmt_dep

        scope.update(get_bindings(statement))


@get_requirements.register(ast.Return)
def _get_requirements_for_return(node):
    """
    ..code:: python

        Return(expr? value)

    """
    if node.value:
        yield from get_requirements(node.value)


@get_requirements.register(ast.Delete)
def _get_requirements_for_delete(node):
    """
    ..code:: python

        Delete(expr* targets)
    """
    for target in node.targets:
        yield from get_requirements(target)


@get_requirements.register(ast.Assign)
def _get_requirements_for_assign(node):
    """
    ..code:: python

        Assign(expr* targets, expr value, string? type_comment)
    """
    yield from get_requirements(node.value)


@get_requirements.register(ast.AugAssign)
def _get_requirements_for_aug_assign(node):
    """
    ..code:: python

        AugAssign(expr target, operator op, expr value)
    """
    yield from get_requirements(node.value)


@get_requirements.register(ast.AnnAssign)
def _get_requirements_for_ann_assign(node):
    """
    ..code:: python

        # 'simple' indicates that we annotate simple name without parens
        AnnAssign(expr target, expr annotation, expr? value, int simple)

    """
    # Can be None for type declaration.
    if node.value:
        yield from get_requirements(node.value)


@get_requirements.register(ast.For)
@get_requirements.register(ast.AsyncFor)
def _get_requirements_for_for(node):
    """
    ..code:: python

        # use 'orelse' because else is a keyword in target languages
        For(
            expr target,
            expr iter,
            stmt* body,
            stmt* orelse,
            string? type_comment,
        )

    ..code:: python

        AsyncFor(
            expr target,
            expr iter,
            stmt* body,
            stmt* orelse,
            string? type_comment,
        )
    """
    bindings = set(get_bindings(node))

    yield from get_requirements(node.iter)

    for stmt in node.body:
        for requirement in get_requirements(stmt):
            if requirement.name not in bindings:
                yield requirement

    for stmt in node.orelse:
        for requirement in get_requirements(stmt):
            if requirement.name not in bindings:
                yield requirement


@get_requirements.register(ast.While)
def _get_requirements_for_while(node):
    """
    ..code:: python

        While(expr test, stmt* body, stmt* orelse)
    """
    yield from get_requirements(node.test)

    for stmt in node.body:
        yield from get_requirements(stmt)

    for stmt in node.orelse:
        yield from get_requirements(stmt)


@get_requirements.register(ast.If)
def _get_requirements_for_if(node):
    """
    ..code:: python

        If(expr test, stmt* body, stmt* orelse)
    """
    yield from get_requirements(node.test)

    for stmt in node.body:
        yield from get_requirements(stmt)

    for stmt in node.orelse:
        yield from get_requirements(stmt)


@get_requirements.register(ast.With)
@get_requirements.register(ast.AsyncWith)
def _get_requirements_for_with(node):
    """
    ..code:: python

        With(withitem* items, stmt* body, string? type_comment)

    ..code:: python

        AsyncWith(withitem* items, stmt* body, string? type_comment)

    .. code:: python

        WithItem(expr context_expr, expr? optional_vars)
    """
    bindings = set(get_bindings(node))

    for item in node.items:
        yield from get_requirements(item.context_expr)

    for stmt in node.body:
        for requirement in get_requirements(stmt):
            if requirement.name not in bindings:
                yield requirement


@get_requirements.register(ast.Raise)
def _get_requirements_for_raise(node):
    """
    ..code:: python

        Raise(expr? exc, expr? cause)
    """
    if node.exc:
        yield from get_requirements(node.exc)

    if node.cause:
        yield from get_requirements(node.cause)


@get_requirements.register(ast.Try)
def _get_requirements_for_try(node):
    """
    ..code:: python

        Try(
            stmt* body,
            excepthandler* handlers,
            stmt* orelse,
            stmt* finalbody,
        )

        ExceptHandler(expr? type, identifier? name, stmt* body)
    """
    for stmt in node.body:
        yield from get_requirements(stmt)

    for handler in node.handlers:
        if handler.type:
            yield from get_requirements(handler.type)

        for stmt in handler.body:
            yield from get_requirements(stmt)

    for stmt in node.orelse:
        yield from get_requirements(stmt)

    for stmt in node.finalbody:
        yield from get_requirements(stmt)


@get_requirements.register(ast.Assert)
def _get_requirements_for_assert(node):
    """
    ..code:: python

        Assert(expr test, expr? msg)

    """
    yield from get_requirements(node.test)

    if node.msg:
        yield from get_requirements(node.msg)


@get_requirements.register(ast.Import)
def _get_requirements_for_import(node):
    """
    ..code:: python

        Import(alias* names)
    """
    return
    yield


@get_requirements.register(ast.ImportFrom)
def _get_requirements_for_import_from(node):
    """
    ..code:: python

        ImportFrom(identifier? module, alias* names, int? level)

    """
    return
    yield


@get_requirements.register(ast.Global)
def _get_requirements_for_global(node):
    """
    ..code:: python

        Global(identifier* names)
    """
    for name in node.names:
        yield Requirement(
            name=name,
            lineno=node.lineno,
            col_offset=node.col_offset,
            scope=Scope.GLOBAL,
        )


@get_requirements.register(ast.Nonlocal)
def _get_requirements_for_non_local(node):
    """
    ..code:: python

        Nonlocal(identifier* names)
    """
    for name in node.names:
        yield Requirement(
            name=name,
            lineno=node.lineno,
            col_offset=node.col_offset,
            scope=Scope.NONLOCAL,
        )


@get_requirements.register(ast.Expr)
def _get_requirements_for_expr(node):
    """
    ..code:: python

        Expr(expr value)
    """
    yield from get_requirements(node.value)


@get_requirements.register(ast.Pass)
@get_requirements.register(ast.Break)
@get_requirements.register(ast.Continue)
def _get_requirements_for_control_flow(node):
    """
    ..code:: python

        Pass | Break | Continue

    """
    return
    yield


@get_requirements.register(ast.BoolOp)
def _get_requirements_for_bool_op(node):
    """
    ..code:: python

        # BoolOp() can use left & right?
        # expr
        BoolOp(boolop op, expr* values)
    """
    for value in node.values:
        yield from get_requirements(value)


@get_requirements.register(ast.NamedExpr)
def _get_requirements_for_named_expr(node):
    """
    ..code:: python

        NamedExpr(expr target, expr value)
    """
    raise NotImplementedError("TODO")


@get_requirements.register(ast.BinOp)
def _get_requirements_for_bin_op(node):

    """
    ..code:: python

        BinOp(expr left, operator op, expr right)
    """
    yield from get_requirements(node.left)
    yield from get_requirements(node.right)


@get_requirements.register(ast.UnaryOp)
def _get_requirements_for_unary_op(node):
    """
    ..code:: python

        UnaryOp(unaryop op, expr operand)
    """

    yield from get_requirements(node.operand)


@get_requirements.register(ast.Lambda)
def _get_requirements_for_lambda(node):
    """
    ..code:: python

        Lambda(arguments args, expr body)
    """
    scope = _get_scope_from_arguments(node.args)
    for requirement in get_requirements(node.body):
        if requirement.name not in scope:
            yield requirement


@get_requirements.register(ast.IfExp)
def _get_requirements_for_if_exp(node):
    """
    ..code:: python

        IfExp(expr test, expr body, expr orelse)
    """
    yield from get_requirements(node.test)
    yield from get_requirements(node.body)
    yield from get_requirements(node.orelse)


@get_requirements.register(ast.Dict)
def _get_requirements_for_dict(node):
    """
    ..code:: python

        Dict(expr* keys, expr* values)
    """
    for key, value in zip(node.keys, node.values):
        yield from get_requirements(value)


@get_requirements.register(ast.Set)
def _get_requirements_for_set(node):
    """
    ..code:: python

        Set(expr* elts)
    """
    for elt in node.elts:
        yield from get_requirements(elt)


@get_requirements.register(ast.ListComp)
def _get_requirements_for_list_comp(node):
    """
    ..code:: python

        ListComp(expr elt, comprehension* generators)
    """
    requirements = []
    requirements += get_requirements(node.elt)

    for generator in node.generators:
        requirements += get_requirements(generator.iter)

        for if_expr in generator.ifs:
            requirements += get_requirements(if_expr)

    bindings = set(get_bindings(node))

    for requirement in requirements:
        if requirement.name not in bindings:
            yield requirement


@get_requirements.register(ast.SetComp)
def _get_requirements_for_set_comp(node):
    """
    ..code:: python

        SetComp(expr elt, comprehension* generators)
    """
    requirements = []
    requirements += get_requirements(node.elt)

    for generator in node.generators:
        requirements += get_requirements(generator.iter)

        for if_expr in generator.ifs:
            requirements += get_requirements(if_expr)

    bindings = set(get_bindings(node))

    for requirement in requirements:
        if requirement.name not in bindings:
            yield requirement


@get_requirements.register(ast.DictComp)
def _get_requirements_for_dict_comp(node):
    """
    ..code:: python

        DictComp(expr key, expr value, comprehension* generators)
    """
    requirements = []
    requirements += get_requirements(node.key)
    requirements += get_requirements(node.value)

    for generator in node.generators:
        requirements += get_requirements(generator.iter)

        for if_expr in generator.ifs:
            requirements += get_requirements(if_expr)

    bindings = set(get_bindings(node))

    for requirement in requirements:
        if requirement.name not in bindings:
            yield requirement


@get_requirements.register(ast.GeneratorExp)
def _get_requirements_for_generator_exp(node):
    """
    ..code:: python

        GeneratorExp(expr elt, comprehension* generators)
    """
    requirements = []
    requirements += get_requirements(node.elt)

    for generator in node.generators:
        requirements += get_requirements(generator.iter)

        for if_expr in generator.ifs:
            requirements += get_requirements(if_expr)

    bindings = set(get_bindings(node))

    for requirement in requirements:
        if requirement.name not in bindings:
            yield requirement


@get_requirements.register(ast.Await)
def _get_requirements_for_await(node):
    """
    ..code:: python

        # the grammar constrains where yield expressions can occur
        Await(expr value)
    """
    yield from get_requirements(node.value)


@get_requirements.register(ast.Yield)
def _get_requirements_for_yield(node):
    """
    ..code:: python

        Yield(expr? value)
    """
    if node.value:
        yield from get_requirements(node.value)


@get_requirements.register(ast.YieldFrom)
def _get_requirements_for_yield_from(node):
    """
    ..code:: python

        YieldFrom(expr value)
    """
    yield from get_requirements(node.value)


@get_requirements.register(ast.Compare)
def _get_requirements_for_compare(node):
    """
    ..code:: python

        # need sequences for compare to distinguish between
        # x < 4 < 3 and (x < 4) < 3
        Compare(expr left, cmpop* ops, expr* comparators)
    """
    yield from get_requirements(node.left)

    for comp in node.comparators:
        yield from get_requirements(comp)


@get_requirements.register(ast.Call)
def _get_requirements_for_call(node):
    """
    ..code:: python

        Call(expr func, expr* args, keyword* keywords)
    """
    yield from get_requirements(node.func)

    for arg in node.args:
        yield from get_requirements(arg)

    for kwarg in node.keywords:
        yield from get_requirements(kwarg.value)


@get_requirements.register(ast.FormattedValue)
def _get_requirements_for_formatted_value(node):
    """
    ..code:: python

        FormattedValue(expr value, int? conversion, expr? format_spec)
    """
    yield from get_requirements(node.value)


@get_requirements.register(ast.JoinedStr)
def _get_requirements_for_joined_str(node):
    """
    ..code:: python

        JoinedStr(expr* values)
    """
    return
    yield


@get_requirements.register(ast.Constant)
def _get_requirements_for_constant(node):
    """
    ..code:: python

        Constant(constant value, string? kind)
    """
    return
    yield


@get_requirements.register(ast.Attribute)
def _get_requirements_for_attribute(node):
    """
    ..code:: python

        # the following expression can appear in assignment context
        Attribute(expr value, identifier attr, expr_context ctx)
    """
    assert isinstance(node.ctx, (ast.Load, ast.Del))
    yield from get_requirements(node.value)


@get_requirements.register(ast.Subscript)
def _get_requirements_for_subscript(node):
    """
    ..code:: python

        Subscript(expr value, expr slice, expr_context ctx)
    """
    assert isinstance(node.ctx, (ast.Load, ast.Del))
    yield from get_requirements(node.value)
    yield from get_requirements(node.slice)


@get_requirements.register(ast.Starred)
def _get_requirements_for_starred(node):
    """
    ..code:: python

        Starred(expr value, expr_context ctx)
    """
    assert isinstance(node.ctx, (ast.Load, ast.Del))
    yield from get_requirements(node.value)


@get_requirements.register(ast.Name)
def _get_requirements_for_name(node):

    """
    ..code:: python

        Name(identifier id, expr_context ctx)
    """
    assert isinstance(node.ctx, (ast.Load, ast.Del))
    yield Requirement(
        name=node.id, lineno=node.lineno, col_offset=node.col_offset
    )


@get_requirements.register(ast.List)
def _get_requirements_for_list(node):

    """
    ..code:: python

        List(expr* elts, expr_context ctx)
    """
    assert isinstance(node.ctx, (ast.Load, ast.Del))
    for element in node.elts:
        yield from get_requirements(element)


@get_requirements.register(ast.Tuple)
def _get_requirements_for_tuple(node):
    """
    ..code:: python

        Tuple(expr* elts, expr_context ctx)

    """
    assert isinstance(node.ctx, (ast.Load, ast.Del))
    for element in node.elts:
        yield from get_requirements(element)


@get_requirements.register(ast.Slice)
def _get_requirements_for_slice(node):
    """
    ..code:: python

        # can appear only in Subscript
        Slice(expr? lower, expr? upper, expr? step)

    """
    if node.lower is not None:
        yield from get_requirements(node.lower)

    if node.upper is not None:
        yield from get_requirements(node.upper)

    if node.step is not None:
        yield from get_requirements(node.step)


if sys.version_info < (3, 9):

    @get_requirements.register(ast.Index)
    def _get_requirements_for_index(node):
        """
        ..code:: python

            # can appear only in Subscript
            Index(expr value)
        """
        yield from get_requirements(node.value)
